from ... import SystemClass
from ...decorators import tools
from ..._formatters import __check_elements
import pandas
from typing import List, Tuple


@tools
def get_all_taps_number(distSys: SystemClass) -> pandas.DataFrame:
    """
    Get the tap number of all regulators.

    Args:
        distSys:  An instance of  [SystemClass][dssdata.SystemClass].
    Returns:
        The tap number of all regulators.
    """
    return get_tap_number(distSys, distSys.dss.RegControls.AllNames())


@tools
def get_tap_number(distSys: SystemClass, names: List[str]) -> pandas.DataFrame:
    """
    Get the tap number of regulators.

    Args:
        distSys : An instance of  [SystemClass][dssdata.SystemClass].
        names : Regulators names

    Returns:
        The tap number of regulators.
    """

    def get_one(reg_name: str) -> Tuple[str, int]:
        distSys.dss.RegControls.Name(reg_name)
        return (reg_name, int(distSys.dss.RegControls.TapNumber()))

    __check_elements(names, distSys.dss.RegControls.AllNames())

    return pandas.DataFrame(
        data=tuple(map(get_one, names)), columns=["reg_name", "tap"]
    )



import pandas
from ... import SystemClass
from ...decorators import tools

from ..._tools import __get_bus_v_pu, __get_bus_ph, __get_bus_ang
from ..._formatters import __identify_ph_config, __check_elements
from typing import List


@tools
def get_all(distSys: SystemClass) -> pandas.DataFrame:
    """
    Get line to neutral voltages (module and angle) in p.u. and phases configuration from buses. See [get_from_buses][dssdata.tools.voltages.get_from_buses].

    Args:
        distSys : An instance of  [SystemClass][dssdata.SystemClass].

    Returns:
        [type]: Line to neutral voltages (module and angle) in p.u. and phases configuration from buses. 
    """  # noqa

    buses = distSys.dss.Circuit.AllBusNames()
    return get_from_buses(distSys, buses)


@tools
def get_from_buses(distSys: SystemClass, buses: List[str]) -> pandas.DataFrame:
    """

    Get line to neutral voltages (module and angle) in p.u. and phases configuration from buses. 

    |    | bus_name | v_pu_a  | v_pu_b  | v_pu_c  | ang_a | ang_b  | ang_c | phases |
    |----|-----------|---------|---------|---------|-------|--------|-------|--------|
    | 0  | sourcebus | 0.99997 | 0.99999 | 0.99995 | 30.0  | -90.0  | 150.0 | abc    |
    | 1  | 646       |   NaN      | 1.01803 | 1.00026 |   NaN    | -122.0 | 117.8 | bc     |
    | 2 | 611       |    NaN     |    NaN     | 0.96083 |   NaN    |    NaN    | 115.7 | c      |
    | 3 | 652       | 0.97533 |    NaN     |   NaN      | -5.3  |    NaN    |  NaN     | a      |


    Args:
        distSys : An instance of  [SystemClass][dssdata.SystemClass].
        buses : Buses names.

    Returns:
        [type]: Line to neutral voltages (module and angle) in p.u. and phases configuration from buses. 
    """  # noqa

    __check_elements(buses, distSys.dss.Circuit.AllBusNames())

    def agreggate(bus: str):
        return (
            __get_bus_ang(distSys, bus),
            __get_bus_v_pu(distSys, bus),
            __identify_ph_config(__get_bus_ph(distSys, bus)),
        )

    (ang_list, v_pu_list, ph_list) = tuple(
        zip(*(agreggate(bus) for bus in buses))
    )

    df_bus_names = pandas.DataFrame(buses, columns=["bus_name"])
    df_v_pu = pandas.DataFrame(
        v_pu_list, columns=["v_pu_a", "v_pu_b", "v_pu_c"]
    )
    df_ang = pandas.DataFrame(ang_list, columns=["ang_a", "ang_b", "ang_c"])
    df_ph = pandas.DataFrame(ph_list, columns=["phase"])

    result = pandas.concat(
        (df_bus_names, df_v_pu, df_ang, df_ph), axis=1, sort=False
    )
    return result



import pandas as pd
from ... import SystemClass
from ...decorators import tools

from ..._formatters import (
    __identify_ph_config,
    __get_mag_vanish,
    __get_ang_vanish,
    __remove_nones_from_lists,
    __check_elements,
)

from typing import List


@tools
def get_infos(distSys: SystemClass, names: List[str]) -> pd.DataFrame:
    """
    Get some relevant infos from lines.
    Ex:

    |    | name   | bus1 | ph_bus1 | bus2 | ph_bus2 | I(A)_bus1_ph_a | I(A)_bus1_ph_b | I(A)_bus1_ph_c | I(A)_bus2_ph_a | I(A)_bus2_ph_b | I(A)_bus2_ph_c | ang_bus1_ph_a | ang_bus1_ph_b | ang_bus1_ph_c | ang_bus2_ph_a | ang_bus2_ph_b | ang_bus2_ph_c | kw_losses | kvar_losses | emergAmps | normAmps | perc_NormAmps | perc_EmergAmps |
    |----|--------|------|---------|------|---------|----------------|----------------|----------------|----------------|----------------|----------------|---------------|---------------|---------------|---------------|---------------|---------------|-----------|-------------|-----------|----------|---------------|----------------|
    | 0  | 650632 | rg60 | abc     | 632  | abc     | 562.609        | 419.029        | 591.793        | 562.61         | 419.03         | 591.794        | -28.7         | -141.3        | 93.4          | 151.3         | 38.7          | -86.6         | 60.737    | 196.015     | 600.0     | 400.0    | 1.479         | 0.986          |
    | 1  | 632670 | 632  | abc     | 670  | abc     | 481.916        | 218.055        | 480.313        | 481.916        | 218.055        | 480.313        | -27.2         | -135.2        | 99.6          | 152.8         | 44.8          | -80.4         | 12.991    | 41.495      | 600.0     | 400.0    | 1.205         | 0.803          |
    | 2  | 670671 | 670  | abc     | 671  | abc     | 473.795        | 188.824        | 424.942        | 473.795        | 188.824        | 424.942        | -27.0         | -132.6        | 101.3         | 153.0         | 47.4          | -78.7         | 22.729    | 72.334      | 600.0     | 400.0    | 1.184         | 0.79           |

    Args:
        distSys : An instance of [SystemClass][dssdata.SystemClass].
        names : Lines names.

    Returns:
        Lines infos.
    """  # noqa

    __check_elements(names, distSys.dss.Lines.AllNames())

    def build_line_dicts(distSys: SystemClass, line_name: str) -> dict:
        def vanish_line_infos(bus_raw: list, current_raw: list) -> tuple:
            bus_name = bus_raw[0]
            phs_raw = list(map(lambda bus: int(bus), bus_raw[1:]))
            phs_data = phs_raw if phs_raw != [] else [1, 2, 3]
            phs = __identify_ph_config(phs_data)
            currents_mag = __get_mag_vanish(phs_data, current_raw)
            currents_ang = __get_ang_vanish(phs_data, current_raw)

            return (bus_name, phs, currents_mag, currents_ang)

        distSys.dss.Lines.Name(line_name)
        losses = distSys.dss.CktElement.Losses()
        normalAmps = distSys.dss.CktElement.NormalAmps()
        emergAmps = distSys.dss.CktElement.EmergAmps()
        currents_raw = distSys.dss.CktElement.CurrentsMagAng()
        currents_raw_bus1 = currents_raw[: int(len(currents_raw) / 2)]
        currents_raw_bus2 = currents_raw[int(len(currents_raw) / 2):]

        bus_raw = distSys.dss.Lines.Bus1().split(".")

        (bus_name1, phs1, currents_mag1, currents_ang1) = vanish_line_infos(
            bus_raw, currents_raw_bus1
        )
        bus_raw = distSys.dss.Lines.Bus2().split(".")
        (bus_name2, phs2, currents_mag2, currents_ang2) = vanish_line_infos(
            bus_raw, currents_raw_bus2
        )

        currents_mag1_calc = __remove_nones_from_lists(currents_mag1)
        currents_mag2_calc = __remove_nones_from_lists(currents_mag2)

        return {
            "name": line_name,
            "bus1": bus_name1,
            "ph_bus1": phs1,
            "bus2": bus_name2,
            "ph_bus2": phs2,
            "I(A)_bus1_ph_a": currents_mag1[0],
            "I(A)_bus1_ph_b": currents_mag1[1],
            "I(A)_bus1_ph_c": currents_mag1[2],
            "I(A)_bus2_ph_a": currents_mag2[0],
            "I(A)_bus2_ph_b": currents_mag2[1],
            "I(A)_bus2_ph_c": currents_mag2[2],
            "ang_bus1_ph_a": currents_ang1[0],
            "ang_bus1_ph_b": currents_ang1[1],
            "ang_bus1_ph_c": currents_ang1[2],
            "ang_bus2_ph_a": currents_ang2[0],
            "ang_bus2_ph_b": currents_ang2[1],
            "ang_bus2_ph_c": currents_ang2[2],
            "kw_losses": losses[0] / 1000,
            "kvar_losses": losses[1] / 1000,
            "emergAmps": emergAmps,
            "normAmps": normalAmps,
            "perc_NormAmps": max(currents_mag1_calc + currents_mag2_calc)
            / normalAmps,
            "perc_EmergAmps": max(currents_mag1_calc + currents_mag2_calc)
            / emergAmps,
        }

    return pd.DataFrame(
        tuple(
            map(lambda line_name: build_line_dicts(distSys, line_name), names,)
        )
    )


@tools
def get_all_infos(distSys: SystemClass) -> pd.DataFrame:
    """
    Get some relevant infos from all lines. See [get_infos][dssdata.tools.lines.get_infos].

    Args:
        distSys: An instance of  [SystemClass][dssdata.SystemClass]

    Returns:
        All lines infos
    """  # noqa
    line_names = distSys.dss.Lines.AllNames()
    return get_infos(distSys, line_names)



from os import getcwd, chdir
from os import path as pathfunc
from typing import Iterable, List

__version__ = "0.1.2"


class SystemClass:
    """
    The distribution system abstraction class.
    """  # noqa: E501

    import opendssdirect

    def __init__(self, *, path: str, kV: Iterable[float], loadmult: float = 1):
        """
        Args:
            path (str): The path to the file that describes the distribution system.
            kV (Iterable[float]): The base voltages. See ```voltagebases``` in [OpenDSS User Manual](http://svn.code.sf.net/p/electricdss/code/trunk/Distrib/Doc/OpenDSSManual.pdf).
            loadmult (float, optional): The load multiplier.  See ```loadmult``` in [OpenDSS User Manual](http://svn.code.sf.net/p/electricdss/code/trunk/Distrib/Doc/> OpenDSSManual.pdf).
        """  # noqa: E501

        with open(path, "rt") as file:
            self._dsscontent = file.read().splitlines()

        self.__path = path
        self.__folder = pathfunc.split(path)[0]
        self.__dss_file = pathfunc.split(path)[1]
        self.__kV = kV
        self.__loadmult = loadmult
        self.init_sys()
        self.__name = self.dss.Circuit.Name()

    @property
    def dss(cls):
        """
        The instance of [OpenDSSDirect.py](https://github.com/dss-extensions/OpenDSSDirect.py).

        """  # noqa: E501

        return cls.opendssdirect

    @property
    def dsscontent(self) -> List[str]:
        """
        The OpenDSS commands of the ```.dss``` file in [path][dssdata.SystemClass.path].

        As well, you can change the list of commands.

        Returns:
            OpenDSS commands.
        """  # noqa: E501
        return self._dsscontent

    @dsscontent.setter
    def dsscontent(self, content: List[str]):
        self._dsscontent = content
        self.init_sys()

    @property
    def name(self) -> str:
        """
        Returns:
            The name of the distribution system.
        """  # noqa: E501
        return self.__name

    @property
    def path(self) -> str:
        """
        Returns:
            The path of the ```.dss``` file.
        """  # noqa: E501
        return self.__path

    @property
    def kV(self) -> Iterable[float]:
        """
        Returns:
            The base voltages. See ```voltagebases``` in [OpenDSS User Manual](http://svn.code.sf.net/p/electricdss/code/trunk/Distrib/Doc/OpenDSSManual.pdf).
        """  # noqa: E501
        return self.__kV

    @property
    def loadmult(self) -> float:
        """
        Returns:
            The load multiplier.  See ```loadmult``` in [OpenDSS User Manual](http://svn.code.sf.net/p/electricdss/code/trunk/Distrib/Doc/OpenDSSManual.pdf)..
        """  # noqa: E501
        return self.__loadmult

    def run_command(self, cmd: str) -> str:
        """
        Run a comand on OpenDSS.

        Args:
            cmd (str): A OpenDSS command.

        Raises:
            Exception: If the command is invalid.
        Returns:
            The OpenDSS command returns
        """  # noqa: E501

        # value = self.dss.run_command(cmd, dss=self.dss)
        self.dss.Text.Command(cmd)
        error = self.error
        if error != "":
            raise Exception(error)

        return self.dss.Text.Result()

    def init_sys(self):
        """
        Run the commands in [dsscontent][dssdata.SystemClass.dsscontent].
        """
        directory = getcwd()
        self.dss.Basic.ClearAll()
        if self.__folder != "":
            chdir(self.__folder)
            list(map(self.run_command, self._dsscontent,))
            chdir(directory)
        else:
            list(map(self.run_command, self._dsscontent,))

        self.run_command(f"Set voltagebases={self.__kV}")
        self.run_command("calcv")
        self.run_command(f"Set loadmult = {self.__loadmult}")

    @property
    def error(cls) -> str:
        """
        Returns:
            The error of the last OpenDSS command.
        """
        return cls.dss.Error.Description()



import pandas as pd
from ... import SystemClass
from ..._formatters import __check_elements


def __build_pd_dicts(
    distSys: SystemClass, element_name: str, element_type: str
) -> dict:

    distSys.dss.PDElements.Name(str(element_type) + "." + str(element_name))
    typ = distSys.dss.CktElement.Name().replace("." + str(element_name), "")

    losses = distSys.dss.CktElement.Losses()

    return {
        "type": typ,
        "name": element_name,
        "kw_losses": losses[0] / 1000,
        "kvar_losses": losses[1] / 1000,
    }


def pd_element_loss(
    distSys: SystemClass, element_name: str, element_type: str
) -> pd.DataFrame:
    """
    Get PD Element loss. 
    Ex:
        
    |   |     type    | name |     kw_losses     |    kvar_losses    |
    |:-:|:-----------:|:----:|:-----------------:|:-----------------:|
    | 0 | Transformer | xfm1 | 5.552671994055243 | 10.09627035828575 |

    Args:
        distSys: An instance of  [SystemClass][dssdata.SystemClass]
        element_name: The name of the desired PD element
        element_type: The type of the PD element (Line or Transformer)

    Returns:
        A DataFrame containing the losses of the desired PD element.
    """  # noqa

    __check_elements(
        list((str(element_type) + "." + str(element_name)).split()),
        distSys.dss.Circuit.AllElementNames(),
    )
    pd_loss = []
    pd_loss.append(__build_pd_dicts(distSys, element_name, element_type))
    return pd.DataFrame(pd_loss)


def pd_element_loss_list(
    distSys: SystemClass, element_names: list, element_type: str
) -> pd.DataFrame:
    """
    Get PD Element loss List. 
    Ex:
        
    |   |     type    | name |      kw_losses      |     kvar_losses     |
    |:-:|:-----------:|:----:|:-------------------:|:-------------------:|
    | 0 | Transformer |  sub | 0.03228776756674051 | 0.26246840671868993 |
    | 1 | Transformer | reg1 | 0.12209426402417012 | 0.12385869008488953 |
    | 2 | Transformer | reg2 | 0.06534502545557916 | 0.06707698704162612 |
    | 3 | Transformer | reg3 |  0.1350894299906213 | 0.13685391995031387 |
    | 4 | Transformer | xfm1 |  5.552671994055243  |  10.09627035828575  |


    Args:
        distSys: An instance of  [SystemClass][dssdata.SystemClass]
        element_names: A list of names of the desired PD elements
        element_type: The type of the PD elements (Line or Transformer)

    Returns:
        A DataFrame containing the losses of the desired list of PD elements.
    """  # noqa

    if element_type == "Line":
        __check_elements(element_names, distSys.dss.Lines.AllNames())
    elif element_type == "Transformer":
        __check_elements(element_names, distSys.dss.Transformers.AllNames())

    return pd.DataFrame(
        tuple(
            map(
                lambda element_name: __build_pd_dicts(
                    distSys, element_name, element_type
                ),
                element_names,
            )
        )
    )


def get_all_line_losses(distSys: SystemClass) -> pd.DataFrame:
    """
    Get all lines losses. 
    Ex:
        
    |    | type |  name  |       kw_losses       |       kvar_losses      |
    |:--:|:----:|:------:|:---------------------:|:----------------------:|
    |  0 | Line | 650632 |   60.73738438443188   |   196.01456922721653   |
    |  1 | Line | 632670 |   12.990633124585496  |    41.49451118066639   |
    |  2 | Line | 670671 |   22.728758590972518  |    72.33414340631373   |
    |  3 | Line | 671680 | 8.613828479544935e-12 |  -0.004169229516017848 |
    |  4 | Line | 632633 |   0.8244871671261499  |   1.0561418323197722   |
    |  5 | Line | 632645 |    2.75857850181032   |   2.4159107795492454   |
    |  6 | Line | 645646 |   0.5274715389783668  |   0.41973513183818434  |
    |  7 | Line | 692675 |   4.1629544212549225  |    2.419339661740261   |
    |  8 | Line | 671684 |   0.5794876384501113  |   0.47068061342113654  |
    |  9 | Line | 684611 |   0.3824044250881998  |   0.38734916932047053  |
    | 10 | Line | 684652 |   0.7998267312559038  |    0.230879175578375   |
    | 11 | Line | 671692 | 9.054614813067019e-06 | 4.3655745685100556e-14 |

    Args:
        distSys: An instance of  [SystemClass][dssdata.SystemClass]

    Returns:
        A DataFrame containing all line losses.
    """  # noqa

    element_type = "Line"

    return pd.DataFrame(
        tuple(
            map(
                lambda element_name: __build_pd_dicts(
                    distSys, element_name, element_type
                ),
                distSys.dss.Lines.AllNames(),
            )
        )
    )


def get_all_transformers_losses(distSys: SystemClass) -> pd.DataFrame:
    """
    Get all transformers losses. 
    Ex:
        
    |   |     type    | name |      kw_losses      |     kvar_losses     |
    |:-:|:-----------:|:----:|:-------------------:|:-------------------:|
    | 0 | Transformer |  sub | 0.03228776756674051 | 0.26246840671868993 |
    | 1 | Transformer | reg1 | 0.12209426402417012 | 0.12385869008488953 |
    | 2 | Transformer | reg2 | 0.06534502545557916 | 0.06707698704162612 |
    | 3 | Transformer | reg3 |  0.1350894299906213 | 0.13685391995031387 |
    | 4 | Transformer | xfm1 |  5.552671994055243  |  10.09627035828575  |

    Args:
        distSys: An instance of  [SystemClass][dssdata.SystemClass]

    Returns:
        A DataFrame containing all transformers losses.
    """  # noqa

    element_type = "Transformer"

    return pd.DataFrame(
        tuple(
            map(
                lambda element_name: __build_pd_dicts(
                    distSys, element_name, element_type
                ),
                distSys.dss.Transformers.AllNames(),
            )
        )
    )


def get_all_pd_elements_losses(distSys: SystemClass) -> pd.DataFrame:
    """
    Get all PD Elements losses. 
    Ex:
        
    |    |     type    |  name  |       kw_losses       |       kvar_losses      |
    |:--:|:-----------:|:------:|:---------------------:|:----------------------:|
    |  0 | Transformer |   sub  |  0.03228776756674051  |   0.26246840671868993  |
    |  1 | Transformer |  reg1  |  0.12209426402417012  |   0.12385869008488953  |
    |  2 | Transformer |  reg2  |  0.06534502545557916  |   0.06707698704162612  |
    |  3 | Transformer |  reg3  |   0.1350894299906213  |   0.13685391995031387  |
    |  4 | Transformer |  xfm1  |   5.552671994055243   |    10.09627035828575   |
    |  5 |     Line    | 650632 |   60.73738438443188   |   196.01456922721653   |
    |  6 |     Line    | 632670 |   12.990633124585496  |    41.49451118066639   |
    |  7 |     Line    | 670671 |   22.728758590972518  |    72.33414340631373   |
    |  8 |     Line    | 671680 | 8.613828479544935e-12 |  -0.004169229516017848 |
    |  9 |     Line    | 632633 |   0.8244871671261499  |   1.0561418323197722   |
    | 10 |     Line    | 632645 |    2.75857850181032   |   2.4159107795492454   |
    | 11 |     Line    | 645646 |   0.5274715389783668  |   0.41973513183818434  |
    | 12 |     Line    | 692675 |   4.1629544212549225  |    2.419339661740261   |
    | 13 |     Line    | 671684 |   0.5794876384501113  |   0.47068061342113654  |
    | 14 |     Line    | 684611 |   0.3824044250881998  |   0.38734916932047053  |
    | 15 |     Line    | 684652 |   0.7998267312559038  |    0.230879175578375   |
    | 16 |     Line    | 671692 | 9.054614813067019e-06 | 4.3655745685100556e-14 |

    Args:
        distSys: An instance of  [SystemClass][dssdata.SystemClass]

    Returns:
        A DataFrame containing all PD Elements losses.
    """  # noqa

    line_losses = get_all_line_losses(distSys)

    transformer_losses = get_all_transformers_losses(distSys)

    return pd.concat([transformer_losses, line_losses])


def get_total_pd_elements_losses(distSys: SystemClass) -> pd.DataFrame:
    """
    Get Total PD Elements losses. 
    Ex:
        
    |   |       name      |   kw_losses_total  | kvar_losses_total |
    |:-:|:---------------:|:------------------:|:-----------------:|
    | 0 | all_pd_elements | 112.39948405966965 | 327.9256193105294 |

    Args:
        distSys: An instance of  [SystemClass][dssdata.SystemClass]

    Returns:
        A DataFrame containing the sum of losses of all PD Elements.
    """  # noqa

    data_loss = []
    data = {
        "name": "all_pd_elements",
        "kw_losses_total": sum(
            get_all_pd_elements_losses(distSys)["kw_losses"]
        ),
        "kvar_losses_total": sum(
            get_all_pd_elements_losses(distSys)["kvar_losses"]
        ),
    }

    data_loss.append(data)
    return pd.DataFrame(data_loss)



import pandas
from .. import SystemClass
from ..decorators import actions, mode
from typing import Iterable, Tuple, Callable


@mode
def run_static_pf(
    distSys: SystemClass,
    actions: Iterable[Callable] = (lambda distSys: None,),
    tools: Iterable[Callable] = (lambda distSys: None,),
) -> tuple:
    """
    Run the static power flow mode.
    To see how it works, see [Learning DSSData](../tutorial/#static-power-flow).

    Args:
        distSys: An instance of [SystemClass][dssdata.SystemClass]
        actions: Actions functions.
        tools: Tools functions.

    Returns:
        Tools functions returns
    """  # noqa
    [action(distSys) for action in actions]
    distSys.run_command("set mode=Snap")
    distSys.dss.Solution.Solve()
    return tuple(tool(distSys) for tool in tools)


@actions
def cfg_tspf(
    distSys: SystemClass, step_size: str = "1h", initial_time: tuple = (0, 0)
) -> None:
    """
    Set the time series mode in the distribution system.

    Args:
        distSys: An instance of [SystemClass][dssdata.SystemClass]
        step_size: The size of step time. See "Stepsize" in [OpenDSS User Manual](http://svn.code.sf.net/p/electricdss/code/trunk/Distrib/Doc/OpenDSSManual.pdf).
        initial_time:  See "Time" in [OpenDSS User Manual](http://svn.code.sf.net/p/electricdss/code/trunk/Distrib/Doc/OpenDSSManual.pdf).
    """  # noqa

    cmd = f"set mode=daily stepsize={step_size} "
    cmd2 = f'time = "{initial_time[0]}, {initial_time[1]}"'
    distSys.run_command(cmd + cmd2)


@mode
def run_tspf(
    distSys: SystemClass,
    num_steps: int,
    actions: Iterable[Callable] = (lambda distSys: None,),
    tools: Iterable[Callable] = (lambda distSys: None,),
) -> Tuple[pandas.DataFrame]:
    """
    Run the time series power flow.
    To see how it works, see [Learning DSSData](../tutorial/#time-series-power-flow).

    Args:
        distSys: An instance of [SystemClass][dssdata.SystemClass]
        num_steps : Number of time steps.
        actions: Actions functions.
        tools: Tools functions.
        

    Returns:
        Tools functions returns for all steps
    """  # noqa

    def runmode_one_step(distSys):
        distSys.dss.Solution.Number(1)
        distSys.dss.Solution.Solve()

    def run_tool(distSys, tool, step):
        df = tool(distSys)
        df["step"] = step
        return df

    def run_concepts(distSys, actions, tools, step):
        for action in actions:
            action(distSys)

        runmode_one_step(distSys)

        return [run_tool(distSys, tool, step) for tool in tools]

    def all_steps_tool(all_steps_result, steps, indx_tool):
        return pandas.concat(
            [all_steps_result[step][indx_tool] for step in steps],
            ignore_index=True,
        )

    steps = range(0, num_steps)

    all_steps_result = [
        run_concepts(distSys, actions, tools, step) for step in steps
    ]

    return [
        all_steps_tool(all_steps_result, steps, indx)
        for indx, _ in enumerate(tools)
    ]



from . import SystemClass
from ._formatters import (
    __get_mag_vanish,
    __get_ang_vanish,
)
from .decorators import tools


@tools
def __get_bus_v_pu_ang(distSys: SystemClass, bus: str):
    distSys.dss.Circuit.SetActiveBus(bus)
    return distSys.dss.Bus.puVmagAngle()


@tools
def __get_bus_ph(distSys: SystemClass, bus: str):
    distSys.dss.Circuit.SetActiveBus(bus)
    return distSys.dss.Bus.Nodes()


@tools
def __get_bus_v_pu(distSys: SystemClass, bus: str):
    v_pu_ang_dss = __get_bus_v_pu_ang(distSys, bus)
    list_ph = __get_bus_ph(distSys, bus)
    v_pu = __get_mag_vanish(list_ph, v_pu_ang_dss)
    return v_pu


@tools
def __get_bus_ang(distSys: SystemClass, bus: str):
    v_pu_ang_dss = __get_bus_v_pu_ang(distSys, bus)
    list_ph = __get_bus_ph(distSys, bus)
    ang = __get_ang_vanish(list_ph, v_pu_ang_dss)
    return ang



from functools import partial
import warnings


def __isodd(value):
    return value % 2 != 0


def __iseven(value):
    return value % 2 == 0


def __aux_vanish(list_ph: list, mag_or_ang: list, indx: int):
    try:
        return mag_or_ang[list_ph.index(indx + 1)]
    except ValueError:
        return None


def __get_mag_vanish(list_ph: list, data: list) -> tuple:
    # The mags is in even indexes
    mag_indexes = filter(__iseven, range(0, len(data)))

    mag_formatted = tuple(
        map(lambda indx: __format_mag(data[indx]), mag_indexes)
    )

    func_aux = partial(__aux_vanish, list_ph, mag_formatted)

    return tuple(map(func_aux, range(0, 3)))


def __get_ang_vanish(list_ph: list, data: list) -> tuple:
    # The angs is in isodd indexes
    ang_indexes = filter(__isodd, range(0, len(data)))

    ang_formatted = tuple(
        map(lambda indx: __format_ang(data[indx]), ang_indexes)
    )

    func_aux = partial(__aux_vanish, list_ph, ang_formatted)

    return tuple(map(func_aux, range(0, 3)))


def __identify_ph_config(phs: list) -> str:
    def aux_ph(ph: int):
        if ph == 1:
            return "a"
        elif ph == 2:
            return "b"
        elif ph == 3:
            return "c"
        else:
            msg = f"The phase {ph} is not a valid number. Only 1, 2 or 3 is aceppted. This phase will be ignored"  # noqa
            warnings.warn(msg, SyntaxWarning)
            return ""

    return "".join(map(aux_ph, phs))


def __format_mag(value: float) -> float:
    return round(value, 4)  # Valor padrão do  OpenDSS


def __format_ang(value: float) -> float:
    return round(value, 2)  # Valor padrão do  OpenDSS


def __remove_nones_from_lists(data: list) -> tuple:
    return tuple(filter(lambda value: value is not None, data))


def __check_elements(element_names: list, all_elements_names: list) -> None:
    """
    Verifica se todos os elementos da lista estão em um outra lista.
    Args:
        element_names: lista menor a ser verificada.
        all_elements_names:lista maior contendo todo o conteúdo
    Returns:
        None
    Raises:
        Exception:  Se algum item de element_names não estiver em
                    all_elements_names
    """  # noqa
    for element_name in element_names:
        if element_name not in all_elements_names:
            raise ValueError(
                f"The element {element_name} is not declared"
                + "on distribution system"
            )



import pandas
from typing import List
from operator import add


def get_taps_changes(tapDataFrame: pandas.DataFrame) -> pandas.DataFrame:
    """
    Count the taps changes. That function is a reduction function.

    Args:
        tapDataFrame: The return of [get_all_taps_number][dssdata.tools.regs.get_all_taps_number] or [get_tap_number][dssdata.tools.regs.get_tap_number].

    Returns:
        The number of taps changes after time series power flow.
    """  # noqa

    def calc_one_step_chgs(first: list, second: list):
        return map(lambda data1, data2: abs(data1 - data2), first, second)

    def sum_changes_taps(list_df: List[list]):
        first = list_df[0]
        second = list_df[1]
        if len(list_df) == 2:
            return calc_one_step_chgs(first, second)
        else:
            return map(
                add,
                calc_one_step_chgs(first, second),
                sum_changes_taps(list_df[1:]),
            )

    iter_by_step = tapDataFrame.groupby(["step"])

    list_df = tuple(map(lambda data: data[1]["tap"].to_list(), iter_by_step))
    reg_names = tuple(
        map(lambda data: data[1]["reg_name"].to_list(), iter_by_step)
    )[0]

    return pandas.DataFrame(
        data=tuple(zip(reg_names, sum_changes_taps(list_df))),
        columns=["reg_name", "number_changes_tap"],
    )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['dssdata',
 'dssdata.decorators',
 'dssdata.pfmodes',
 'dssdata.reductions',
 'dssdata.reductions.regs',
 'dssdata.tools',
 'dssdata.tools.lines',
 'dssdata.tools.losses',
 'dssdata.tools.regs',
 'dssdata.tools.voltages']

package_data = \
{'': ['*']}

install_requires = \
['OpenDSSDirect.py', 'pandas']

setup_kwargs = {
    'name': 'dssdata',
    'version': '0.1.5',
    'description': 'Organizing OpenDSS data',
    'long_description': "# DSSData\n\n\n_**⚡A python micro-framework for steady-state simulation and data analysis of electrical distribution systems modeled on [OpenDSS](https://www.epri.com/#/pages/sa/opendss?lang=en).**_\n\nMode support: Static and Time-series.\n\n## Why DSSData?\n\nThe propose of DSSData is to facilitate the steady-state simulation of modern electrical distribution systems, such as microgrid, smart grids, and smart cities.\n\nWith DSSData you can easily make your own super new fancy operation strategies with storage or generators, probabilistic simulation, or a simple impact studies of a distributed generator. See an example in our [Tutorial](https://felipemarkson.github.io/dssdata/tutorial/).\n\n**_All you need is your base distribution system modeled in OpenDSS!!!_**\n\n### Easy to simulate\n\nWe built the DSSData for you just write what you want in a simple function, plugin on a power flow mode, and run. \n\nYou don't need anymore write a routine to run each power flow per time. \n\n## Documentation\n\nSee [DSSData Documentation](https://felipemarkson.github.io/dssdata).\n\n## Installation\n\nWe strongly recommend the use of virtual environments manager.\n\n### Using pip\n\n```console\npip install dssdata\n```\n\n### Using poetry\n\n```console\npoetry add dssdata\n```\n\n## Help us to improve DSSData\n\nSee our [Issue](https://github.com/felipemarkson/dssdata/issues) section!\n\n\n## Contributors: \n\n- [JonasVil](https://github.com/felipemarkson/power-flow-analysis/commits?author=JonasVil)\n",
    'author': 'Felipe M. S. Monteiro',
    'author_email': 'fmarkson@outlook.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/felipemarkson/dssdata',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from functools import wraps
from typing import Callable


def tools(func: Callable) -> Callable:
    """
    See [Creating your first tool](/tutorial/#creating-your-first-tool).

    Args:
        func:  A tool function

    Returns:
        The return of the tool function
    """

    @wraps(func)
    def wrapper(distSys, *args, **kwds):
        if distSys.dss.Circuit.Name() != distSys.name:
            distSys.init_sys()
        data = func(distSys, *args, **kwds)
        return data

    return wrapper


def actions(func: Callable) -> Callable:
    """
    See [Creating your first action](/tutorial/#creating-your-first-action).

    Args:
        func:  A action function
    """

    @wraps(func)
    def wrapper(distSys, *args, **kwds):
        if distSys.dss.Circuit.Name() != distSys.name:
            distSys.init_sys()
        func(distSys, *args, **kwds)

    return wrapper


def mode(func: Callable) -> Callable:
    @wraps(func)
    def wrapper(distSys, *args, **kwds):
        if distSys.dss.Circuit.Name() != distSys.name:
            distSys.init_sys()
        data = func(distSys, *args, **kwds)
        distSys.init_sys()
        return data

    return wrapper
